#pragma once

#include "regularizer.h"
#include "soft_constraints.h"
#include "squared_distance.h"
#include "cross_correlation_sphere.h"
#include "cross_correlation_cube.h"
#include "gradient_squared_distance.h"
#include "landmarks.h"
#include "mutual_information.h"
#include "sub_function.h"
#include "unary_function.h"
